# Automatically generated with Copilot 

import polars as pl
import kra  

def test_label_expr_encode():
    df = pl.DataFrame({"animal": ["cat", "dog", "cat"]})
    result = df.with_columns(pl.col("animal").label.encode().alias("encoded"))
    # The encoded column should be [0, 1, 0] or [1, 0, 1] depending on categorical order, but unique values
    encoded = result["encoded"].to_list()
    assert set(encoded) == {0, 1}
    assert encoded[0] == encoded[2]
    assert encoded[0] != encoded[1]

def test_label_series_encode():
    s = pl.Series(["cat", "dog", "cat"])
    encoded = s.label.encode()
    # The encoded series should be [0, 1, 0] or [1, 0, 1] depending on categorical order, but unique values
    values = encoded.to_list()
    assert set(values) == {0, 1}
    assert values[0] == values[2]
    assert values[0] !=


# Automatically generated with Copilot 
import polars as pl
import pytest
import kra  

def test_apply_lowercase():
    df = pl.DataFrame({"A": [1], "B": [2]})
    result = df.cols.apply(lambda x: x.lower())
    assert result.columns == ["a", "b"]

def test_to_lowercase():
    df = pl.DataFrame({"A": [1], "B": [2]})
    result = df.cols.to_lowercase()
    assert result.columns == ["a", "b"]

def test_to_uppercase():
    df = pl.DataFrame({"a": [1], "b": [2]})
    result = df.cols.to_uppercase()
    assert result.columns == ["A", "B"]

def test_to_titlecase():
    df = pl.DataFrame({"first name": [1], "last name": [2]})
    result = df.cols.to_titlecase()
    assert result.columns == ["First Name", "Last Name"]

def test_to_camelcalse():
    df = pl.DataFrame({"first name": [1], "last name": [2]})
    result = df.cols.to_camelcalse()
    assert result.columns == ["FirstName", "LastName"]

def test_to_snakecase():
    df = pl.DataFrame({"First Name": [1], "Last Name": [2]})
    result = df.cols.to_snakecase()
    assert result.columns == ["first_name", "last_name"]

def test_replace():
    df = pl.DataFrame({"foo-bar": [1], "baz-bar": [2]})
    result = df.cols.replace("-bar", "_suffix")
    assert result.columns == ["foo_suffix", "baz_suffix"]

def test_has_all_true():
    df = pl.DataFrame({"a": [1], "b": [2]})
    has, missing = df.cols.has_all(["a", "b"])
    assert has is True
    assert missing == []

def test_has_all_false():
    df = pl.DataFrame({"a": [1], "b": [2]})
    has, missing = df.cols.has_all(["a", "c"])
    assert has is False
    assert missing == ["c"] or set(missing) == {"c"}

def test_has_any_true():
    df = pl.DataFrame({"a": [1], "b": [2]})
    assert df.cols.has_any(["b", "c"]) is True

def test_has_any_false():
    df = pl.DataFrame({"a": [1], "b": [2]})
    assert df.cols.has_any(["x", "y"]) is False

def test_has_exactly_true():
    df = pl.DataFrame({"a": [1], "b": [2]})
    assert df.cols.has_exactly(["a", "b"]) is True

def test_has_exactly_false():
    df = pl.DataFrame({"a": [1], "b": [2]})
    assert df.cols.has_exactly(["a"]) is False

def test_rename_partial():
    df = pl.DataFrame({"a": [1], "b": [2]})
    result = df.cols.rename({"a": "x", "c": "y"})
    assert result.columns == ["x", "b"]


import polars as pl

from kra.utils import Cloneable


@pl.api.register_expr_namespace('label')
class LabelExpr:
    """
    Expression namespace for label encoding in polars expressions.

    Example
    -------
    >>> import polars as pl
    >>> import kra  # noqa: F401, needed for registration
    >>> df = pl.DataFrame({"animal": ["cat", "dog", "cat"]})
    >>> df.with_columns(pl.col("animal").label.encode().alias("encoded"))
    shape: (3, 2)
    ┌────────┬─────────┐
    │ animal ┆ encoded │
    ├────────┼─────────┤
    │ cat    ┆ 0       │
    │ dog    ┆ 1       │
    │ cat    ┆ 0       │
    └────────┴─────────┘
    """

    def __init__(self, expr: pl.Expr) -> None:
        self._expr = expr
    
    def encode(self):
        """
        Encode string values as integer labels using categorical encoding.

        Returns
        -------
        pl.Expr
            An expression that encodes string values as integer labels.

        Examples
        --------
        >>> import polars as pl
        >>> import kra  # noqa: F401
        >>> df = pl.DataFrame({"animal": ["cat", "dog", "cat"]})
        >>> df.with_columns(pl.col("animal").label.encode().alias("encoded"))
        shape: (3, 2)
        ┌────────┬─────────┐
        │ animal ┆ encoded │
        ├────────┼─────────┤
        │ cat    ┆ 0       │
        │ dog    ┆ 1       │
        │ cat    ┆ 0       │
        └────────┴─────────┘
        """
        return self._expr.cast(pl.String).cast(pl.Categorical).to_physical()


@pl.api.register_series_namespace('label')
class LabelSeries:
    """
    Series namespace for label encoding in polars Series.

    Example
    -------
    >>> import polars as pl
    >>> import kra  # noqa: F401, needed for registration
    >>> s = pl.Series(["cat", "dog", "cat"])
    >>> s.label.encode()
    shape: (3,)
    Series: '' [u32]
    [
        0
        1
        0
    ]
    """

    def __init__(self, series: pl.Series) -> None:
        self._series = series
    
    def encode(self):
        """
        Encode string values as integer labels using categorical encoding.

        Returns
        -------
        pl.Series
            A Series with integer labels for each unique string value.

        Examples
        --------
        >>> import polars as pl
        >>> import kra  # noqa: F401
        >>> s = pl.Series(["cat", "dog", "cat"])
        >>> s.label.encode()
        shape: (3,)
        Series: '' [u32]
        [
            0
            1
            0
        ]
        """
        return self._series.cast(pl.String).cast(pl.Categorical).to_physical()
        # d = {k: v for v, k in enumerate(self._series.unique())}
        # return self._series.replace(d, return_dtype=pl.Int32)
        



# Automatically generated with Copilot 

import polars as pl
import kra 

def test_drop_null_cols_removes_null_column():
    df = pl.DataFrame({"a": [1, 2], "b": [None, None]})
    result = df.drop_null_cols()
    assert result.columns == ["a"]
    assert "b" not in result.columns

def test_drop_null_cols_keeps_non_null():
    df = pl.DataFrame({"a": [1, 2], "b": [3, 4]})
    result = df.drop_null_cols()
    assert result.columns == ["a", "b"]

def test_fork_creates_multiple_dataframes():
    df = pl.DataFrame({"a": [1, 2]})
    forks = df.fork([{"b": [10, 20]}, {"c": [100, 200]}])
    assert len(forks) == 2
    assert forks[0].columns == ["a", "b"]
    assert forks[1].columns == ["a", "c"]
    assert forks[0]["b"].to_list() == [10, 20]
    assert forks[1]["c"].to_list()


# Automatically generated with Copilot

import polars as pl
import numpy as np
import kra  # noqa: F401

def test_to_dod():
    df = pl.DataFrame({"id": [1, 2], "name": ["Alice", "Bob"]})
    result = df.to_dod("id")
    assert result == {1: {'id': 1, 'name': 'Alice'}, 2: {'id': 2, 'name': 'Bob'}}

def test_from_dod():
    dod = {1: {"name": "Alice"}, 2: {"name": "Bob"}}
    df = pl.from_dod(dod, "id")
    assert set(df.columns) == {"id", "name"}
    assert set(df["id"].to_list()) == {1, 2}
    assert set(df["name"].to_list()) == {"Alice", "Bob"}

def test_from_arraylike():
    data = np.array([[1, 2, 3], [4, 5, 6]])
    df = pl.from_arraylike(data, schema=["a", "b"], orient="col")
    assert df.shape == (3, 2)
    assert set(df.columns) == {"a", "b"}
    assert df["a"].to_list() == [1, 2, 3]
    assert df["b"].to_list() == [4, 5, 6]

def test_from_memoryview():
    data = np.array([[1, 2], [3, 4]])
    mv = memoryview(data)
    df = pl.from_memoryview(mv, schema=["x", "y"], orient="col")
    assert set(df.columns) == {"x", "y"}
    assert df["x"].to_list() == [1, 3]
    assert df["y"].to_list() == [2, 4]

def test_maybe_col_exists():
    df = pl.DataFrame({"a": [1, 2]})
    result = df.select(kra.maybe_col("a", 0))
    assert result["a"].to_list() == [1, 2]

def test_maybe_col_missing():
    df = pl.DataFrame({"a": [1, 2]})
    result = df.select(kra.maybe_col("b", 0))
    assert result["b"].to_list() == [0, 0]

def test_split_entries_by():
    df = pl.DataFrame({"a": [1, 2], "n": [2, 3]})
    result = df.split_entries_by("n")
    assert set(result.columns) == {"a", "n"}
    assert result.shape[0] == 5

def test_to_dicts():
    df = pl.DataFrame({"g": ["a", "a", "b"], "x": [1, 2, 3]})
    d = df.groupby("g").to_dicts()
    assert set(d.keys()) == {"a", "b"}
    assert all(isinstance(v, list) for v in d.values())
    assert d["a"][0]["g"] == "a"

def test_to_set():
    s = pl.Series([1, 2, 2, 3])
    result = s.to_set()
    assert result



from kra.polars_api import extend_polars, extend_polars_dataframe
from kra.utils import from_dod, to_dod, from_dict_of_dicts, to_dict_of_dicts
from kra.columns import Cols
from kra.process import drop_null_cols
from kra.label import LabelSecris, LabelExpr

import kra.kra
from kra.kra import encode_labels


import polars as pl
from kra.utils import extend_polars_dataframe


@extend_polars_dataframe
def drop_null_cols(df: pl.DataFrame) -> pl.DataFrame:
    """
    Exclude columns of type Null from the DataFrame.

    Returns
    -------
    pl.DataFrame
        DataFrame with all columns of type Null removed.

    Examples
    --------
    >>> import polars as pl
    >>> import kra 
    >>> df = pl.DataFrame({"a": [1, 2], "b": [None, None]})
    >>> df.drop_null_cols()
    shape: (2, 1)
    ┌─────┐
    │ a   │
    ├─────┤
    │ 1   │
    │ 2   │
    └─────┘
    """
    return df.with_columns(pl.exclude(pl.Null))

@extend_polars_dataframe
def fork(df: pl.DataFrame, new_dfs: list) -> list[pl.DataFrame]:
    """
    Fork a DataFrame into multiple new DataFrames with additional columns.

    Parameters
    ----------
    new_dfs : list of dict
        Each dict specifies new columns to add to a forked DataFrame.

    Returns
    -------
    list of pl.DataFrame
        List of new DataFrames, each with the specified additional columns.

    Examples
    --------
    >>> import polars as pl
    >>> import kra  
    >>> df = pl.DataFrame({"a": [1, 2]})
    >>> forks = df.fork([{"b": [10, 20]}, {"c": [100, 200]}])
    >>> for f in forks:
    ...     print(f)
    shape: (2, 2)
    ┌─────┬─────┐
    │ a   ┆ b   │
    ├─────┼─────┤
    │ 1   ┆ 10  │
    │ 2   ┆ 20  │
    └─────┴─────┘
    shape: (2, 2)
    ┌─────┬───────┐
    │ a   ┆ c     │
    ├─────┼───────┤
    │ 1   ┆ 100   │
    │ 2   ┆ 200   │
    └─────┴───────┘
    """
    # TODO: 
    return [df.with_columns(**n) for n in new_dfs]



import polars as pl
import numpy as np
import numpy.typing as npt

import polars.dataframe.group_by as plg 
# import polars.convert as plt

from kra.polars_api import extend_polars, extend_polars_dataframe, extend_polars_group_by, extend_polars_series


class Cloneable:

    def __init__(self, df: pl.DataFrame) -> None:
        self._df = df

    def _clone_if(self, in_place) -> pl.DataFrame:
        if not in_place:
            df = self._df.clone()
        else:
            df = self._df
        return df


@extend_polars_dataframe
def to_dod(df: pl.DataFrame, key: str) -> dict:
    """
    Convert a polars DataFrame to a dict of dicts, using a column as the dictionary key.

    Parameters
    ----------
    df : pl.DataFrame
        The DataFrame to convert.
    key : str
        The column name to use as the key for the outer dictionary.
        Each value in this column must be unique.

    Returns
    -------
    dict
        Dictionary mapping unique values from the specified column to row dictionaries.

    Examples
    --------
    >>> import polars as pl
    >>> import kra  # noqa: F401
    >>> df = pl.DataFrame({"id": [1, 2], "name": ["Alice", "Bob"]})
    >>> df.to_dod("id")
    {1: {'id': 1, 'name': 'Alice'}, 2: {'id': 2, 'name': 'Bob'}}
    """
    return {row[key]: row for row in df.to_dicts()}

to_dict_of_dicts = to_dod
extend_polars_dataframe(to_dict_of_dicts)


@extend_polars
def from_dod(dod: dict, column: str, **kwargs) -> pl.DataFrame:
    """
    Construct a DataFrame from a dictionary of dictionaries.

    Parameters
    ----------
    dod : dict
        Dictionary of dictionaries, where each key-value pair represents a row.
    column : str
        The name of the column to use for the keys of the outer dictionary.
    **kwargs
        Additional keyword arguments passed to `pl.from_dicts`.

    Returns
    -------
    pl.DataFrame
        DataFrame constructed from the dict of dicts.

    Examples
    --------
    >>> import polars as pl
    >>> import kra  # noqa: F401
    >>> dod = {1: {"name": "Alice"}, 2: {"name": "Bob"}}
    >>> pl.from_dod(dod, "id")
    shape: (2, 2)
    ┌─────┬───────┐
    │ id  ┆ name  │
    ├─────┼───────┤
    │ 1   ┆ Alice │
    │ 2   ┆ Bob   │
    └─────┴───────┘
    """
    return pl.from_dicts([d | {column: k} for k, d in dod.items()], **kwargs)

from_dict_of_dicts = from_dod
extend_polars(from_dict_of_dicts)


@extend_polars
def from_arraylike(data: npt.ArrayLike,
    schema = None,*,
    schema_overrides: None = None,
    orient = None,
) -> pl.DataFrame:
    """Construct a DataFrame from a Cython memoryview or any other Array-like. This operation clones data.
    
    Parameters
    ----------
    data : :class:`npt.ArrayLike`
        Array-like data
    schema : Sequence of str, (str,DataType) pairs, or a {str:DataType,} dict
        The DataFrame schema may be declared in several ways:

        * As a dict of {name:type} pairs; if type is None, it will be auto-inferred.
        * As a list of column names; in this case types are automatically inferred.
        * As a list of (name,type) pairs; this is equivalent to the dictionary form.

        If you supply a list of column names that does not match the names in the
        underlying data, the names given here will overwrite them. The number
        of names given in the schema should match the underlying data dimensions.
    schema_overrides : dict, default None
        Support type specification or override of one or more columns; note that
        any dtypes inferred from the columns param will be overridden.
    orient : {None, 'col', 'row'}
        Whether to interpret two-dimensional data as columns or as rows. If None,
        the orientation is inferred by matching the columns and data dimensions. If
        this does not yield conclusive results, column orientation is used.

    Returns
    -------
    DataFrame

    Examples
    --------
    >>> import numpy as np
    >>> data = np.array([[1, 2, 3], [4, 5, 6]])
    >>> df = pl.from_numpy(data, schema=["a", "b"], orient="col")
    >>> df
    shape: (3, 2)
    ┌─────┬─────┐
    │ a   ┆ b   │
    │ --- ┆ --- │
    │ i64 ┆ i64 │
    ╞═════╪═════╡
    │ 1   ┆ 4   │
    │ 2   ┆ 5   │
    │ 3   ┆ 6   │
    └─────┴─────┴
    
    """
    return pl.from_numpy(np.asarray(data), schema, schema_overrides, orient)

from_memoryview = from_arraylike
extend_polars(from_memoryview)


@extend_polars
def maybe_col(name, default=None) -> pl.Expr:
    """
    Return a column expression for a column with the given name, or a default value if the column is missing.

    Parameters
    ----------
    name : str
        The column name to select.
    default : Any, optional
        The default value to use if the column is missing.

    Returns
    -------
    pl.Expr
        A polars expression selecting the column or the default value.

    Examples
    --------
    >>> import polars as pl
    >>> import kra
    >>> df = pl.DataFrame({"a": [1, 2]})
    >>> df.select(kra.maybe_col("a", 0))
    shape: (2, 1)
    ┌─────┐
    │ a   │
    ├─────┤
    │ 1   │
    │ 2   │
    └─────┘
    >>> df.select(kra.maybe_col("b", 0))
    shape: (2, 1)
    ┌─────┐
    │ b   │
    ├─────┤
    │ 0   │
    │ 0   │
    └─────┘
    """
    # https://github.com/pola-rs/polars/issues/18372
    col = pl.col(f"^{name}$")
    if not default:
        return col
    expr = pl.struct(col, default).struct[0]
    return expr.alias(name)

@extend_polars_dataframe
def split_entries_by(df: pl.DataFrame, column: str) -> pl.DataFrame:
    """
    Repeat and flatten all columns by the values in a given column.

    Parameters
    ----------
    column : str
        The column whose values determine the number of repetitions for each row.

    Returns
    -------
    pl.DataFrame
        DataFrame with rows repeated and flattened according to the column.

    Examples
    --------
    >>> import polars as pl
    >>> import kra  # noqa: F401
    >>> df = pl.DataFrame({"a": [1, 2], "n": [2, 3]})
    >>> df.split_entries_by("n")
    shape: (5, 2)
    ┌─────┬─────┐
    │ a   ┆ n   │
    ├─────┼─────┤
    │ 1   ┆ 1   │
    │ 1   ┆ 1   │
    │ 2   ┆ 1   │
    │ 2   ┆ 1   │
    │ 2   ┆ 1   │
    └─────┴─────┘
    """
    return df.select(pl.all().repeat_by(column).flatten(column)) \
        .with_columns(pl.lit(1).alias(column))


@extend_polars_group_by
def to_dicts(grouped_df: plg.GroupBy) -> dict:
    """
    Convert a polars GroupBy object to a dictionary of lists of dicts.

    Returns
    -------
    dict
        Dictionary mapping group keys to lists of row dicts.

    Examples
    --------
    >>> import polars as pl
    >>> import kra  # noqa: F401
    >>> df = pl.DataFrame({"g": ["a", "a", "b"], "x": [1, 2, 3]})
    >>> df.groupby("g").to_dicts()
    {'a': [{'g': 'a', 'x': 1}, {'g': 'a', 'x': 2}], 'b': [{'g': 'b', 'x': 3}]}
    """
    return { k: v.to_dicts() for k, v in grouped_df}


@extend_polars_series
def to_set(self: pl.Series) -> set:
    """
    Convert a Series to a set of unique values.

    Returns
    -------
    set
        Set of unique values in the Series.

    Examples
    --------
    >>> import polars as pl
    >>> import kra  # noqa: F401
    >>> s = pl.Series([1, 2, 2, 3])
    >>> s.to_set()
    {1, 2, 3}
    """
    return set(self.to_list())



import polars as pl
import polars.dataframe.group_by as plg 

# TODO: add optional name argument

def extend_polars_dataframe(fun: callable):
    """Decorator to extend polars DataFrame API"""
    setattr(pl.DataFrame, fun.__name__, fun)
    return fun # Needed?

def extend_polars_group_by(fun: callable):
    """Decorator to extend polars DataFrame API"""
    setattr(plg.GroupBy, fun.__name__, fun)
    return fun # Needed?

def extend_polars(fun: callable):
    """Decorator to extend polars DataFrame API"""
    setattr(pl, fun.__name__, fun)
    return fun

def extend_polars_series(fun: callable):
    """Decorator to extend polars Series API"""
    setattr(pl.Series, fun.__name__, fun)
    return fun 



import polars as pl
import re

from kra.utils import Cloneable


@pl.api.register_dataframe_namespace('cols')
class Cols(Cloneable):
    """
    A namespace for convenient DataFrame column operations.

    Access via `df.cols` for any polars DataFrame.

    Provides methods for:
      - Bulk renaming and transforming column names (case, pattern, etc.)
      - Checking for presence of columns
      - Safe renaming with partial mappings

    Example
    --------
    >>> import polars as pl
    >>> import kra
    >>> df = pl.DataFrame({"First Name": [1], "Last Name": [2]})
    >>> df.cols.to_snakecase()
    shape: (1, 2)
    ┌────────────┬───────────┐
    │ first_name ┆ last_name │
    └────────────┴───────────┘
    """

    def apply(self, fun: callable, in_place: bool = False):
        """
        Apply a function to all column names.

        Parameters
        ----------
        fun : callable
            Function to apply to each column name.
        in_place : bool, default False
            If True, modify the DataFrame in place. If False, return a copy.

        Returns
        -------
        pl.DataFrame
            DataFrame with transformed column names.

        Examples
        --------
        >>> df = pl.DataFrame({"A": [1], "B": [2]})
        >>> df.cols.apply(lambda x: x.lower())
        shape: (1, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        └─────┴─────┘
        """
        df = self._clone_if(in_place)
        df.columns = [fun(x) for x in df.columns]
        return df 
    
    def to_lowercase(self, in_place: bool = False):
        """
        Convert all column names to lowercase.

        Parameters
        ----------
        in_place : bool, default False
            If True, modify in place. If False, return a copy.

        Returns
        -------
        pl.DataFrame
            DataFrame with lowercase column names.

        Examples
        --------
        >>> df = pl.DataFrame({"A": [1], "B": [2]})
        >>> df.cols.to_lowercase()
        shape: (1, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        └─────┴─────┘
        """
        return self.apply(lambda x: x.lower(), in_place=in_place)
    
    def to_uppercase(self, in_place: bool = False):
        """
        Convert all column names to uppercase.

        Parameters
        ----------
        in_place : bool, default False
            If True, modify in place. If False, return a copy.

        Returns
        -------
        pl.DataFrame
            DataFrame with uppercase column names.

        Examples
        --------
        >>> df = pl.DataFrame({"a": [1], "b": [2]})
        >>> df.cols.to_uppercase()
        shape: (1, 2)
        ┌─────┬─────┐
        │ A   ┆ B   │
        └─────┴─────┘
        """
        return self.apply(lambda x: x.upper(), in_place=in_place)
    
    def to_titlecase(self, in_place: bool = False):
        """
        Convert all column names to title case.

        Parameters
        ----------
        in_place : bool, default False
            If True, modify in place. If False, return a copy.

        Returns
        -------
        pl.DataFrame
            DataFrame with title-cased column names.

        Examples
        --------
        >>> df = pl.DataFrame({"first name": [1], "last name": [2]})
        >>> df.cols.to_titlecase()
        shape: (1, 2)
        ┌────────────┬───────────┐
        │ First Name ┆ Last Name │
        └────────────┴───────────┘
        """
        return self.apply(lambda x: x.title(), in_place=in_place)
    
    def to_camelcalse(self, in_place: bool = False):
        """
        Convert all column names to camel case.

        Parameters
        ----------
        in_place : bool, default False
            If True, modify in place. If False, return a copy.

        Returns
        -------
        pl.DataFrame
            DataFrame with camelCase column names.

        Examples
        --------
        >>> df = pl.DataFrame({"first name": [1], "last name": [2]})
        >>> df.cols.to_camelcalse()
        shape: (1, 2)
        ┌──────────┬─────────┐
        │ FirstName┆ LastName│
        └──────────┴─────────┘
        """
        fun = lambda x: ''.join(word for word in x.title() if not x.isspace())
        return self.apply(fun, in_place=in_place)

    def to_snakecase(self, in_place: bool = False):
        """
        Convert all column names to snake_case.

        Parameters
        ----------
        in_place : bool, default False
            If True, modify in place. If False, return a copy.

        Returns
        -------
        pl.DataFrame
            DataFrame with snake_case column names.

        Examples
        --------
        >>> df = pl.DataFrame({"First Name": [1], "Last Name": [2]})
        >>> df.cols.to_snakecase()
        shape: (1, 2)
        ┌────────────┬───────────┐
        │ first_name ┆ last_name │
        └────────────┴───────────┘
        """
        fun = lambda x: re.sub('\\s', '', re.sub(r'(?<!^)(?=[A-Z])', '_', x)).lower()
        return self.apply(fun, in_place=in_place)
    
    def replace(self, pattern: str | re.Pattern, repl: str, in_place: bool = False):
        """
        Replace a regex pattern in all column names.

        Parameters
        ----------
        pattern : str or re.Pattern
            Pattern to search for.
        repl : str
            Replacement string.
        in_place : bool, default False
            If True, modify in place. If False, return a copy.

        Returns
        -------
        pl.DataFrame
            DataFrame with replaced column names.

        Examples
        --------
        >>> df = pl.DataFrame({"foo-bar": [1], "baz-bar": [2]})
        >>> df.cols.replace("-bar", "_suffix")
        shape: (1, 2)
        ┌──────────┬────────────┐
        │ foo_suffix ┆ baz_suffix │
        └──────────┴────────────┘
        """
        return self.apply(lambda x: re.sub(pattern, repl, x), in_place=in_place)
    
    def has_all(self, columns: list[str], return_missing: bool = True) -> bool | tuple[bool, list[str]]:
        """
        Check if all specified columns are present in the DataFrame.

        Parameters
        ----------
        columns : list of str
            List of column names to check.
        return_missing : bool, default True
            If True, also return a list of missing columns.

        Returns
        -------
        bool or (bool, list of str)
            True if all columns are present, otherwise False.
            If return_missing is True, also returns a list of missing columns.

        Examples
        --------
        >>> df = pl.DataFrame({"a": [1], "b": [2]})
        >>> df.cols.has_all(["a", "b"])
        (True, [])
        >>> df.cols.has_all(["a", "c"])
        (False, ['c'])
        """
        missing = self._missing_cols(columns)
        if return_missing:
            return len(missing) == 0, missing
        return len(missing) == 0
    
    def has_any(self, columns: list[str]) -> bool:
        """
        Check if any of the specified columns are present in the DataFrame.

        Parameters
        ----------
        columns : list of str
            List of column names to check.

        Returns
        -------
        bool
            True if any column is present, otherwise False.

        Examples
        --------
        >>> df = pl.DataFrame({"a": [1], "b": [2]})
        >>> df.cols.has_any(["b", "c"])
        True
        >>> df.cols.has_any(["x", "y"])
        False
        """
        return len(self._common_cols(columns)) > 0
    
    def has_exactly(self, columns: list[str]) -> bool:
        """
        Check if the DataFrame has exactly the specified columns (no more, no less).

        Parameters
        ----------
        columns : list of str
            List of column names to check.

        Returns
        -------
        bool
            True if the DataFrame has exactly these columns, otherwise False.

        Examples
        --------
        >>> df = pl.DataFrame({"a": [1], "b": [2]})
        >>> df.cols.has_exactly(["a", "b"])
        True
        >>> df.cols.has_exactly(["a"])
        False
        """
        return set(columns) == set(self._df.columns)
    
    def rename(self, mapping: dict[str, str]) -> pl.DataFrame:
        """
        Rename columns using a mapping, skipping keys not present in the DataFrame.
        A non-strict version of DataFrame.rename() method.
        It skips missing keys without raising an error.

        Parameters
        ----------
        mapping : dict of str to str
            Mapping from old column names to new names.

        Returns
        -------
        pl.DataFrame
            DataFrame with renamed columns.

        Examples
        --------
        >>> df = pl.DataFrame({"a": [1], "b": [2]})
        >>> df.cols.rename({"a": "x", "c": "y"})
        shape: (1, 2)
        ┌─────┬─────┐
        │ x   ┆ b   │
        └─────┴─────┘
        """
        mapping = {k:v for k, v in mapping.items() if k in self._df.columns}
        return self._df.rename(mapping)
                 
    def _common_cols(self, columns: list[str]):
        return set(columns).intersection(set(self._df.columns))

    def _missing_cols(self, columns: list[str]):
        return set(columns).difference(set(self._df.columns))